"""
  Classes representing the different parts of a BPMN file.
"""

from .    import xml
from .xml import IdentifiedElement

class Flow(IdentifiedElement):
  __tag__ = "bpmn:sequenceFlow"
  __labeled__ = False

  def __init__(self, source=None, target=None):
    super().__init__()
    self._source = source
    self._target = target
    if self._source:
      self._source.outgoing.append(Outgoing(self))
    if self._target:
      self._target.incoming.append(Incoming(self))

  @property
  def source(self):
    if self._source:
      return self._source
    elif self["sourceRef"]:
      obj = self.root.find("id", self["sourceRef"])
      if obj: return obj
    raise Exception(f"no source found on {self}")

  @property
  def target(self):
    if self._target:
      return self._target
    elif self["targetRef"]:
      obj = self.root.find("id", self["targetRef"])
      if obj: return obj
    raise Exception(f"no target found on {self}")

  def __getitem__(self, name):
    if name == "id":
      return f"flow_{self.source['id']}_{self.target['id']}"
    return super().__getitem__(name)

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    attributes.update({
      "id"       : self["id"],
      "sourceRef": self.source["id"],
      "targetRef": self.target["id"]      
    })
    return attributes

class MessageFlow(Flow):
  __tag__ = "bpmn:messageFlow"

class Incoming(xml.Element):
  __tag__ = "bpmn:incoming"
  
  def __init__(self, flow=None):
    super().__init__()
    self.flow = flow
    if self.flow:
      self.text = self.flow["id"]

class Outgoing(xml.Element):
  __tag__ = "bpmn:outgoing"

  def __init__(self, flow=None):
    super().__init__()
    self.flow = flow
    if self.flow:
      self.text = self.flow["id"]

  @property
  def target(self):
    return self.flow.target

class Element(IdentifiedElement):
  __labeled__ = True

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.incoming = []
    self.outgoing = []
    self.x      = 0
    self.y      = 0
    self.width  = 100
    self.height = 80

  def append(self, child):
    if type(child) == Incoming:
      self.incoming.append(child)
      child._parent = self
    elif type(child) == Outgoing:
      self.outgoing.append(child)
      child._parent = self
    else:
      raise ValueError("Element expects only incoming or outgoing flows")
    child._parent = self
    return self

  @property
  def children(self):
    children = []
    if self.incoming:
      children.extend([ flow for flow in self.incoming ])
    if self.outgoing:
      children.extend([ flow for flow in self.outgoing ])
    return children

class MessageEventDefinition(IdentifiedElement):
  __tag__ = "bpmn:messageEventDefinition"

class Event(Element):
  __labeled__ = False

  def __init__(self, name=None, message=False, **kwargs):
    super().__init__(**kwargs)
    self.width   = 36
    self.height  = 36
    if name:
      self["name"] = name
    self.message = message

  def append(self, child):
    if type(child) == MessageEventDefinition:
      self.message = True
    else:
      super().append(child) # something else
    return self

  @property
  def children(self):
    children = super().children
    if self.message:
      children.append(MessageEventDefinition(id=f"MessageEventDefinition_{self['id']}"))
    return children

class Start(Event):
  __tag__    = "bpmn:startEvent"

class IntermediateThrow(Event):
  __tag__ = "bpmn:intermediateThrowEvent"

class IntermediateCatch(Event):
  __tag__ = "bpmn:intermediateCatchEvent"

class End(Event):
  __tag__ = "bpmn:endEvent"

class Task(Element):
  __tag__     = "bpmn:task"

  def __init__(self, name="", **kwargs):
    super().__init__(**kwargs)
    self["name"] = name

  @property
  def process(self):
    return self._parent

  @property
  def lane(self):
    for lane in self.process.laneset.lanes:
      for ref in lane.refs:
        if ref.ref == self:
          return lane
    return None

class UserTask(Task):
  __tag__ = "bpmn:userTask"

class ScriptTask(Task):
  __tag__ = "bpmn:scriptTask"

class ServiceTask(Task):
  __tag__ = "bpmn:serviceTask"

class FlowNodeRef(xml.Element):
  __tag__ = "bpmn:flowNodeRef"

  def __init__(self, ref=None, **kwargs):
    super().__init__(**kwargs)
    self._ref = ref

  @property
  def process(self):
    return self._parent.process

  @property
  def ref(self):
    if self._ref:
      return self._ref
    try:
      return self.process.element(super().text)
    except TypeError:
      pass
    return None
    
  @xml.Element.text.getter
  def text(self):
    # if we have referenced object (possibly resolved from #text)
    if self.ref:
      ref = self.ref.id
    else:
      ref = super().text    # else we should have a textual
    return ref

class Lane(IdentifiedElement):
  __tag__ = "bpmn:lane"
  __horizontal__ = True

  def __init__(self, name=None, *args, **kwargs):
    super().__init__(*args, **kwargs)
    if name:
      self["name"] = name
    self.refs    = []
    self.x      = 30
    self.y      = 0
    self.height = 125
    self.width  = 570

  @property
  def process(self):
    return self._parent.process

  @property
  def elements(self):
    # deref references
    return [ ref.ref for ref in self.refs if ref.ref ] # FIXME: ref.ref is None

  def append(self, child):
    if isinstance(child, FlowNodeRef):       # "native" ref
      self.refs.append(child)
      child._parent = self
    elif isinstance(child, Element):
      wrapped = FlowNodeRef(child)
      child._parent = wrapped
      wrapped._parent = self
      self.refs.append(wrapped)  # wrap it
    else:
      super().append(child) # something else
    return self

  @property
  def children(self):
    return super().children + self.refs

  def has_child(self, child):
    for ref in self.refs:
      if ref.ref == child:
        return True
    return False

class LaneSet(IdentifiedElement):
  __tag__ = "bpmn:laneSet"

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.lanes = []
  
  @property
  def process(self):
    return self._parent
  
  def append(self, child):
    if isinstance(child, Lane):
      self.lanes.append(child)
      child._parent = self
    else:
      super().append(child)
    return self
  
  def __len__(self):
    return len(self.lanes)

  @property
  def children(self):
    return super().children + self.lanes

  def lane_of(self, child):
    for lane in self.lanes:
      if lane.has_child(child):
        return lane
    return None

class Process(IdentifiedElement):
  __tag__ = "bpmn:process"
  
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.laneset = LaneSet(id=f"LaneSet_{self['id']}")
    self._elements = []
  
  def element(self, id):
    for element in self._elements:
      if element["id"] == id:
        return element
    return None
  
  @property
  def participant(self):
    for collaboration in self._parent.collaborations:
      participant = collaboration.find("processRef", self.id)
      if participant:
        return participant
    return None
  
  @property
  def elements(self):
    elems = self._elements  # direct elements
    # add elements that are "hidden" in lanes
    if len(self.laneset) > 0:
      for lane in self.laneset.lanes:
        for element in lane.elements:
          if not element in elems:
            elems.append(element)
    return elems
  
  def append(self, child):
    if isinstance(child, Element):
      self._elements.append(child)
      child._parent = self
    elif isinstance(child, LaneSet):
      self.laneset = child
      child._parent = self
    elif isinstance(child, Lane):
      self.laneset.append(child)
      child._parent = self
    else:
      super().append(child) # generic child handling
    return self

  @property
  def children(self):
    children = super().children.copy()
    children.extend(self.elements)
    if len(self.laneset) > 0:
      children.append(self.laneset)
    return children



import xmltodict
import json
from pathlib import Path

from bpmn_tools.notation      import Definitions
from bpmn_tools.collaboration import Collaboration, Participant
from bpmn_tools.flow          import Process, Start, End, Task
from bpmn_tools.flow          import Flow, FlowNodeRef, Lane, LaneSet
from bpmn_tools.diagrams      import Diagram, Plane, Shape, Edge

from bpmn_tools.layout        import simple

from bpmn_tools.xml           import Element

from bpmn_tools.util import compare

def compare_with_roundtrip(obj1, expected):
  # obj -> xml
  d1 = obj1.as_dict(with_tag=True)
  print("*" * 20, "round 1" , "*" * 20)
  compare(d1, expected)
  
  # xml -> obj
  
  obj2 = Definitions.from_dict(d1)
  
  d2 = obj2.as_dict(with_tag=True)
  print("*" * 20, "round 2" , "*" * 20)
  compare(d2, expected)
  
  return obj2

def test_flow_node_ref():
  a = Task("Some task")
  r1 = FlowNodeRef(a)

  assert r1.ref  == a
  assert r1.text == a.id

  r2 = compare_with_roundtrip(r1, { "bpmn:flowNodeRef" : a.id })
  
  assert r2.ref  == None
  assert r2.text == a.id

def test_lane():
  activities = [
    Start(id="start"),
    Task('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  l1 = Lane("some lane", id="testlane").extend(activities)

  compare_with_roundtrip(l1, {
    "bpmn:lane": {
      "@id": "testlane",
      "@name": "some lane",
      "bpmn:flowNodeRef": [
        "start",
        "hello",
        "end"
      ]
    }
  })

def test_laneset():
  l1 = Lane("lane 1", id="lane1").append(Task("task 1", id="task1"))
  l2 = Lane("lane 2", id="lane2").append(Task("task 2", id="task2"))
  ls = LaneSet(id="laneset").extend([l1, l2])
  
  assert len(ls) == 2

  compare_with_roundtrip(ls,{
    "bpmn:laneSet": {
      "@id": "laneset",
      "bpmn:lane": [
        {
          "@id": "lane1",
          "@name": "lane 1",
          "bpmn:flowNodeRef": "task1"
        },
        {
          "@id": "lane2",
          "@name": "lane 2",
          "bpmn:flowNodeRef": "task2"
        }
      ]
    }
  })

def test_process_with_default_laneset():
  l1 = Lane("lane 1", id="lane1").append(Task("task 1", id="task1"))
  l2 = Lane("lane 2", id="lane2").append(Task("task 2", id="task2"))
  ls = LaneSet(id="laneset").extend([l1, l2])

  process = Process(id="process").append(ls)
  
  compare_with_roundtrip(process, {
    "bpmn:process": {
      "@id": "process",
      "bpmn:task": [
        {
          "@id": "task1",
          "@name": "task 1"
        },
        {
          "@id": "task2",
          "@name": "task 2"
        }
      ],
      "bpmn:laneSet": {
        "@id": "laneset",
        "bpmn:lane": [
          {
            "@id": "lane1",
            "@name": "lane 1",
            "bpmn:flowNodeRef": "task1"
          },
          {
            "@id": "lane2",
            "@name": "lane 2",
            "bpmn:flowNodeRef": "task2"
          }
        ]
      }
    }
  })

def test_diagram_with_lane():
  with open(Path(__file__).resolve().parent / "hello-lanes.bpmn") as fp:
    expected = xmltodict.parse(fp.read())

  activities_lane1 = [
    Start(id="start"),
    Task('Say "Hello!"', id="hello"),
    End(id="end")
  ]

  activities_lane2 = [
    Task('Hear "Hello!"', id="hear")
  ]

  lane2 = Lane("lane 2", id="lane2").extend(activities_lane2)

  process_lane1 = Process(id="process1").extend(activities_lane1).extend([
    Flow(source=activities_lane1[0], target=activities_lane1[1]),
    Flow(source=activities_lane1[1], target=activities_lane1[2])
  ])

  process_lane2 = Process(id="process2").append(lane2)

  collaboration = Collaboration(id="collaboration").extend([
    Participant("participant1", process_lane1, id="participant1"),
    Participant("participant2", process_lane2, id="participant2")
  ])

  definitions = Definitions(id="definitions").extend([
    process_lane1,
    process_lane2,
    collaboration
  ])

  definitions.append(
    Diagram(
      id="diagram",
      plane=Plane(id="plane", element=collaboration)
    )
  )

  simple.layout(definitions)

  compare(definitions.as_dict(with_tag=True), expected)



def Red(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#831311",
    "bioc:fill"              : "#ffcdd2",
    "color:background-color" : "#ffcdd2",
    "color:border-color"     : "#831311"
  }
  return element

def Green(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#205022",
    "bioc:fill"              : "#c8e6c9",
    "color:background-color" : "#c8e6c9",
    "color:border-color"     : "#205022"
  }
  return element

def Orange(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#6b3c00",
    "bioc:fill"              : "#ffe0b2",
    "color:background-color" : "#ffe0b2",
    "color:border-color"     : "#6b3c00"
  }
  return element

def Blue(element):
  element.__color_scheme__ = {
    "bioc:stroke"            : "#0d4372",
    "bioc:fill"              : "#bbdefb",
    "color:background-color" : "#bbdefb",
    "color:border-color"     : "#0d4372"
  }
  return element



from pathlib import Path
import xmltodict
import json

from bpmn_tools.xml     import Element
from bpmn_tools.visitor import Visitor

def test_visiting_hello():
  with open(Path(__file__).resolve().parent / "hello.bpmn") as fp:
    xml = xmltodict.parse(fp.read())
    tree = Element.from_dict(xml)

    visitations = []

    class MyVisitor(Visitor):
      def visit(self, visited):
        if visited["id"]:
          visitations.append(visited["id"])

    tree.accept(MyVisitor())

    assert visitations == [
      "definitions",
      "collaboration",
      "participant",
      "process",
      "start",
      "hello",
      "end",
      "flow_start_hello",
      "flow_hello_end",
      "diagram",
      "plane_collaboration",
      "shape_participant",
      "shape_start",
      "shape_hello",
      "shape_end",
      "edge_flow_start_hello",
      "edge_flow_hello_end"
    ]



from pathlib import Path
import xmltodict

from bpmn_tools.notation  import Definitions
from bpmn_tools.util      import compare

def compare_with_roundtrip(xml1):
  # xml -> obj
  d1 = xmltodict.parse(xml1)
  obj1 = Definitions.from_dict(d1)

  # obj -> xml
  d1g = obj1.as_dict(with_tag=True)
  
  # ensure that generated dict equals original parsed dict
  compare(d1g, d1)
  
  xml2 = xmltodict.unparse(d1g)

  # xml -> obj
  d2 = xmltodict.parse(xml2)
  
  compare(d2, d1)
  
  obj2 = Definitions.from_dict(d2)
  d2g = obj2.as_dict(with_tag=True)
  
  compare(d2g, d1)

def test_hello():
  with open(Path(__file__).resolve().parent / "hello.bpmn") as fp:
    compare_with_roundtrip(fp.read())

def test_hello_lanes():
  with open(Path(__file__).resolve().parent / "hello-lanes.bpmn") as fp:
    compare_with_roundtrip(fp.read())

def test_hello_lanes_with_message():
  with open(Path(__file__).resolve().parent / "hello-lanes-with-message.bpmn") as fp:
    compare_with_roundtrip(fp.read())

def test_hello_colors():
  with open(Path(__file__).resolve().parent / "hello-colors.bpmn") as fp:
    compare_with_roundtrip(fp.read())

def test_message_events():
  with open(Path(__file__).resolve().parent / "message-events.bpmn") as fp:
    compare_with_roundtrip(fp.read())



# decorator based visitor, inspired by https://stackoverflow.com/a/28398903

def _qualname(obj):
  """Get the fully-qualified name of an object (including module)."""
  return obj.__module__ + '.' + obj.__qualname__

def _declaring_class(obj):
  """Get the name of the class that declared an object."""
  name = _qualname(obj)
  return name[:name.rfind('.')]

_methods = {}

def _visitor_impl(self, arg):
  """Actual visitor method implementation."""
  try:
    method = _methods[(_qualname(type(self)), type(arg))]
    return method(self, arg)
  except KeyError:
    pass
  return Visitor.visit(self, arg) 

def visiting(*arg_types):
  """Decorator that creates a visitor method."""
  def decorator(fn):
    declaring_class = _declaring_class(fn)
    for arg_type in arg_types:
      _methods[(declaring_class, arg_type)] = fn
    return _visitor_impl
  return decorator

class Visitor:
  def __init__(self):
    self.depth = -1

  def __enter__(self):
    self.depth += 1
    return self

  def __exit__(self, exc_type, exc_value, exc_tb):
    self.depth -= 1

  def visit(self, visited):
    pass

class PrintingVisitor(Visitor):
  def visit(self, visited):
    print(f"{'   '*self.depth}{visited}")



"""
  Classes representing the different parts of a BPMN file.
"""

from .util import prune

from . import xml

from .collaboration import Participant
from .flow          import Process, Element, Flow, MessageFlow

class Bounds(xml.Element):
  __tag__ = "dc:Bounds"
  
  def __init__(self, x=0, y=0, height=0, width=0):
    super().__init__()
    self["x"]      = str(int(x))
    self["y"]      = str(int(y))
    self["height"] = str(int(height))
    self["width"]  = str(int(width))

class Label(xml.Element):
  __tag__ = "bpmndi:BPMNLabel"
  
  def __init__(self, label=None):
    super().__init__()
    self.text = label

class Shape(xml.Element):
  __tag__ = "bpmndi:BPMNShape"

  def __init__(self, element=None, id=None):
    super().__init__()
    self._element = element
    self.label    = None
    self._bounds  = None

  def append(self, child):
    if isinstance(child, Bounds):
      self._bounds = child
      child._parent = self
    else:
      super().append(child)
    return self

  @property
  def id(self):
    return f"shape_{self.element['id']}"

  @property
  def element(self):
    if self._element:
      return self._element
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  @property
  def attributes(self):
    attributes = {}
    if self.element:
      attributes.update({
        "id"          : self.id,
        "bpmnElement" : self.element["id"]
      })
      if self.element.__horizontal__:
        attributes["isHorizontal"] = "true"
      if self.element.__color_scheme__:
        attributes.update(self.element.__color_scheme__)
    return attributes

  @property
  def children(self):
    children = []
    if self.element:
      children = [
        Bounds(
          x      = self.element.x,
          y      = self.element.y,
          height = self.element.height,
          width  = self.element.width
        )
      ]
      if self.element.__labeled__:
        children.append(Label(self.label))
    return children

class WayPoint(xml.Element):
  __tag__ = "di:waypoint"
  
  def __init__(self, x=0, y=0):
    super().__init__()
    self["x"] = str(int(x))
    self["y"] = str(int(y))

class Edge(xml.Element):
  __tag__ = "bpmndi:BPMNEdge"

  def __init__(self, flow=None, id=None):
    super().__init__()
    self._flow = flow

  @property
  def flow(self):
    if self._flow:
      return self._flow
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  def __getitem__(self, name):
    if name == "id":
      return f"edge_{self.flow['id']}"
    return super().__getitem__(name)

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.flow:
      attributes.update({
        "id"          : self["id"],
        "bpmnElement" : self.flow["id"]
      })
    return attributes

  @property
  def children(self):
    children = super().children.copy()
    if self.flow:
      below = self.flow.target.y > (self.flow.source.y + self.flow.source.height)
      above = (self.flow.target.y + self.flow.target.height) < self.flow.source.y
      if not (above or below): 
        children = [
          WayPoint(
            x=self.flow.source.x + self.flow.source.width,
            y=self.flow.source.y + int(self.flow.source.height/2)
          ),
          WayPoint(
            x=self.flow.target.x,
            y=self.flow.target.y + int(self.flow.target.height/2)
          )
        ]
      else:
        if self.flow.source.y < self.flow.target.y:
          top     = self.flow.source
          bottom  = self.flow.target
          reverse = False
        else:
          top     = self.flow.target
          bottom  = self.flow.source
          reverse = True
        half_way_dist = int((top.y + top.height - bottom.y) / 2)
        children = [
          WayPoint(
            x=top.x + int(top.width/2),
            y=top.y + top.height
          ),
          WayPoint(
            x=top.x + int(top.width/2),
            y=top.y + top.height - half_way_dist
          ),
          WayPoint(
            x=bottom.x + int(bottom.width/2),
            y=bottom.y + half_way_dist
          ),
          WayPoint(
            x=bottom.x + int(bottom.width/2),
            y=bottom.y
          )
        ]
        if reverse:
          children.reverse()
    return children

class Plane(xml.Element):
  __tag__ = "bpmndi:BPMNPlane"

  def __init__(self, id="plane", element=None):
    super().__init__()
    self._element = element
    self["id"] = id
    self._shapes = []
    self._edges  = []

  @property
  def element(self):
    if self._element:
      return self._element
    if self["bpmnElement"]:
      return self.root.find("id", self["bpmnElement"], skip=self)
    return None

  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.element:
      attributes.update({
        "bpmnElement" : self.element["id"],
        "id" : f"plane_{self.element['id']}"
      })
    return attributes

  def append(self, child):
    if isinstance(child, Shape):
      self._shapes.append(child)
      child._parent = self
    elif isinstance(child, Edge):
      self._edges.append(child)
      child._parent = self
    else:
      super().append(child)
    return self

  @property
  def children(self):
    children = super().children.copy()
    if self.element:
      for participant in self.element.children_oftype(Participant):
        children.append(Shape(participant))
        if participant.process:
          for lane in participant.process.laneset.lanes:
            children.append(Shape(lane))
          for element in participant.process.children_oftype(Element):
            children.append(Shape(element))
          for flow in participant.process.children_oftype(Flow):
            children.append(Edge(flow))
      for flow in self.element.children_oftype(MessageFlow):
        children.append(Edge(flow))
    return children

class Diagram(xml.Element):
  __tag__ = "bpmndi:BPMNDiagram"

  def __init__(self, id="diagram", plane=None):
    super().__init__()
    self._plane = plane
    self["id"]  = id

  @property
  def plane(self):
    if self._plane:
      return self._plane
    if self._children:
      return self._children[0]
    return Plane()

  @property
  def children(self):
    return [ self.plane ]



"""

Class that accepts an xmltodict-style dictionary, builds a generic Element
hierarchy and can reproduce the dictionary.

The class can be overridden to create classes dealing with specialized Elements.

A Visitor allows for accessing the entire Element-hierarchy.

"""

import random
import string


class Element():
  __tag__     = "Element"

  def __init__(self, **kwargs):
    self._children   = []
    self._parent     = None
    self._attributes = {}
    self._text       = None

  @property
  def text(self):
    return self._text
  
  @text.setter
  def text(self, value):
    self._text = value

  def __repr__(self):
    label = f"{self.__class__.__name__}"
    if self.attributes:
      label += f"({self.attributes})"
    if self.text:
      label += f" : {self.text}"
    return label

  def __setitem__(self, name, value):
    self._attributes[name] = value

  def __getitem__(self, name):
    try:
      return self._attributes[name]
    except KeyError:
      return None

  def __getattr__(self, name):
    return self[name]

  @property
  def root(self):
    if self._parent:
      return self._parent.root
    else:
      return self

  def find(self, key, value, skip=None, stack=None):
    if stack is None:
      stack = []

    if self in stack:
      logger.warn("avoided recursion")
      return None

    if self is skip:
      return None

    # do I have the key=value attribute?
    try:
      if self._attributes[key] == value:
        return self
    except KeyError:
      pass
    
    # recurse down children
    for child in self.children:
      match = child.find(key, value, skip=skip, stack=stack+[self])
      if match:
        return match

    return None

  def append(self, child):
    self._children.append(child)
    child._parent = self
    return self

  def extend(self, children):
    for child in children:
      self.append(child)
    return self

  @property
  def attributes(self):
    return self._attributes

  @property
  def children(self):
    return self._children

  def children_oftype(self, cls):
    return [
      child for child in self.children \
      if isinstance(child, cls) or isinstance(child.wrapped, cls)
    ]

  def as_dict(self, with_tag=False):
    # collect attributes
    definition = {
      f"@{key}" : value for key, value in self.attributes.items()
    }

    # collect text
    if self.text:
      definition["#text"] = self.text

    # collect children
    for child in self.children:
      d = child.as_dict()
      try:
        definition[child.__tag__].append(d)
      except KeyError:
        definition[child.__tag__] = d
      except AttributeError:
        definition[child.__tag__] = [ definition[child.__tag__] , d ]

    # prune text-only tag
    if list(definition.keys()) == [ "#text" ]:
      definition = definition["#text"]

    # prune empty tag
    if definition == {}:
      definition = None

    if with_tag:
      return { self.__tag__ : definition }
    else:
      return definition
  
  @staticmethod
  def mapped_class(tag, classes):
    if classes:
      for clazz in classes:
        try:
          if clazz.__tag__ == tag:
            return clazz
        except AttributeError:
          pass
    return Element
  
  @classmethod
  def from_dict(cls, d, classes=None, depth=0):
    element_type, element_definition = list(d.items())[0]
    element_class = cls.mapped_class(element_type, classes)
    element = element_class()
    element.__tag__ = element_type
    
    if type(element_definition) == str:
      element_definition = { "#text" : element_definition }
    
    for key, defintions in element_definition.items():
      if key[0] == "@":
        element._attributes[key[1:]] = defintions
      elif key == "#text":
        element.text = defintions
      else:
        if type(defintions) != list:
          defintions = [ defintions ]
        for definition in defintions:
          if definition is None:
            definition = {}
          elif type(definition) is str:
            definition = { "#text" : definition }
          child = Element.from_dict({ key : definition }, classes=classes, depth=depth+1 )
          assert child != element
          element.append(child)

    return element

  def accept(self, visitor):
    with visitor:
      visitor.visit(self)
      for child in self.children:
        child.accept(visitor)

class IdentifiedElement(Element):
  def __init__(self, id=None, **kwargs):
    super().__init__(**kwargs)
    if id is None:
      random_str = ''.join(random.choices(string.ascii_uppercase + string.digits, k=8))
      id = f"{self.__class__.__name__.lower()}_{random_str}"
    self["id"] = id



"""
  Classes representing the different parts of a BPMN file.
"""

from .util import prune

from . import xml

class Participant(xml.Element):
  __tag__        = "bpmn:participant"
  __horizontal__ = True

  def __init__(self, name="participant", process=None, id="participant"):
    super().__init__()
    self["id"]   = id
    self["name"] = name
    self._process = process
    self.x      = 0
    self.y      = 0
    self.height = 125
    self.width  = 600
  
  @property
  def process(self):
    if self._process:
      return self._process
    elif self["processRef"]:
      return self.root.find("id", self["processRef"])
    return None
  
  @property
  def attributes(self):
    attributes = super().attributes.copy()
    if self.process:
      attributes.update({
        "processRef" : self.process["id"]
      })
    return attributes

class Collaboration(xml.Element):
  __tag__ = "bpmn:collaboration"

  def __init__(self, id="collaboration"):
    super().__init__()
    self["id"] = id



from difflib import unified_diff
import json
import xmltodict

def prune(lst):
  if len(lst) == 1:
    return lst[0]
  return lst

def show_diff(result, expected):
  def as_dict(obj):
    return obj.as_dict()
  result   = json.dumps(result,   indent=2, sort_keys=True, default=as_dict)
  expected = json.dumps(expected, indent=2, sort_keys=True, default=as_dict)
  diff = unified_diff(expected.splitlines(keepends=True),
                      result.splitlines(keepends=True),
                      fromfile="expected", tofile="result")
  print("".join(diff), end="")

def compare(result, expected):
  print("=" * 20, "result dict", "=" * 20)
  print(json.dumps(result, indent=2))
  print("=" * 20, "result XML", "=" * 20)
  print(xmltodict.unparse(result, pretty=True))
  print("=" * 20, "diff", "=" * 20)
  show_diff(result, expected)
  assert result == expected



from pathlib import Path
import xmltodict
import json

from bpmn_tools.xml  import Element
from bpmn_tools.util import compare

def test_hello_round_trip():
  with open(Path(__file__).resolve().parent / "hello.bpmn") as fp:
    expected = xmltodict.parse(fp.read())
  tree = Element.from_dict(expected)
  result = tree.as_dict(with_tag=True)
  compare(result, expected)
